import os
import pickle
from typing import Optional

import pandas as pd

from lyzr.base.llm import LiteLLM, LyzrLLMFactory
from lyzr.base.prompt import LyzrPromptFactory


def read_file(
    filepath: str, encoding: Optional[str] = "utf-8", **kwargs
) -> pd.DataFrame:
    if not os.path.exists(filepath):
        raise ValueError(
            f"File '{filepath}' not found. Please provide a valid filepath."
        )
    file_extension = filepath.split(".")[-1]
    try:
        if file_extension == "csv":
            return pd.read_csv(filepath, encoding=encoding, **kwargs)
        elif file_extension == "tsv":
            return pd.read_csv(filepath, sep="\t", encoding=encoding, **kwargs)
        elif file_extension == "txt":
            with open(filepath, "r") as f:
                return f.read(encoding=encoding, **kwargs)
        elif file_extension == "json":
            return pd.read_json(filepath, encoding=encoding, **kwargs)
        elif file_extension in ["xlsx", "xls"]:
            return pd.read_excel(filepath, encoding=encoding, **kwargs)
        elif file_extension == "pkl":
            with open(filepath, "rb") as f:
                return pickle.load(f, encoding=encoding, **kwargs)
        else:
            raise ValueError(
                f"File extension '{file_extension}' not supported. Please provide a csv or pkl file."
            )
    except FileNotFoundError:
        raise FileNotFoundError(
            f"File '{filepath}' not found. Please provide a valid filepath."
        )
    except UnicodeDecodeError:
        raise UnicodeDecodeError(
            f"File '{filepath}' could not be decoded. Please provide a file with utf-8 encoding."
            "If the file is not encoded in utf-8, please provide the encoding as a parameter: file_kwargs={'encoding': 'utf-8'}"
        )


def describe_dataset(
    model: Optional[LiteLLM] = None,
    df: Optional[pd.DataFrame] = None,
    api_key: Optional[str] = None,
    model_type: Optional[str] = None,
    model_name: Optional[str] = None,
) -> str:
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Please provide a valid pandas DataFrame.")

    if model is None:
        model = LyzrLLMFactory(
            api_key=api_key,
            api_type=model_type,
            model=model_name,
        )

    messages = [
        LyzrPromptFactory(
            name="dataset_description", prompt_type="system"
        ).get_message(),
        LyzrPromptFactory(name="dataset_description", prompt_type="user").get_message(
            headers=df.columns.tolist(), df_sample=df.head()
        ),
    ]
    output = model.run(messages=messages)

    return output.message.content



# standart-library imports
import io
import re
import string
import logging

# third-party imports
import numpy as np
import pandas as pd


def get_columns_names(
    df_columns: pd.DataFrame.columns,
    arguments: dict = {},
    columns: list = None,
    logger: logging.Logger = None,
) -> list:
    if isinstance(df_columns, pd.MultiIndex):
        df_columns = df_columns.levels[0]
    if columns is None:
        if "columns" not in arguments or not isinstance(arguments.get("columns"), list):
            return df_columns.to_list()
        columns = arguments.get("columns", [])
    if len(columns) == 0:
        return df_columns.to_list()
    columns_dict = {remove_punctuation_from_string(col): col for col in df_columns}
    column_names = []
    for col in columns:
        if remove_punctuation_from_string(col) not in columns_dict:
            logger.warning(
                "Invalid column name provided: {}. Skipping this column.".format(col)
            )
        else:
            column_names.append(columns_dict[remove_punctuation_from_string(col)])
    return column_names


def remove_punctuation_from_string(value: str) -> str:
    value = str(value).strip()
    value = value.translate(str.maketrans("", "", string.punctuation))
    value = value.replace(" ", "").lower()
    return value


def flatten_list(lst: list):
    for el in lst:
        if isinstance(el, list):
            yield from flatten_list(el)


def _remove_punctuation_from_numeric_string(value) -> str:
    if not isinstance(value, str):
        return value
    negative = False
    value = value.strip()
    if value[0] == "-":
        negative = True
    cleaned = re.sub(r"[^\d.]", "", str(value))
    if cleaned.replace(".", "").isdigit():
        return "-" + cleaned if negative else cleaned
    else:
        return value


def convert_to_numeric(df: pd.DataFrame, columns: list[str]) -> pd.DataFrame:
    for col in columns:
        try:
            df = df.dropna(subset=[col])
            df.loc[:, col] = df.loc[:, col].apply(
                _remove_punctuation_from_numeric_string
            )
            df.loc[:, col] = pd.to_numeric(df.loc[:, col])
            df.loc[:, col] = df.loc[:, col].astype("float")
        except Exception:
            pass
    return df.infer_objects()


def get_info_dict_from_df_dict(df_dict: dict[pd.DataFrame]) -> dict[str]:
    df_info_dict = {}
    for name, df in df_dict.items():
        if not isinstance(df, pd.DataFrame):
            continue
        buffer = io.StringIO()
        df.info(buf=buffer)
        df_info_dict[name] = buffer.getvalue()
    return df_info_dict


def format_df_details(df_dict: dict[pd.DataFrame], df_info_dict: dict[str]) -> str:
    str_output = []
    for name, df in df_dict.items():
        var_name = name.lower().replace(" ", "_")
        if name in df_info_dict and isinstance(df, pd.DataFrame):
            str_output.append(
                f"Dataframe: `{var_name}`\nOutput of `{var_name}.head()`:\n{df.head()}\n\nOutput of `{var_name}.info()`:\n{df_info_dict[name]}\n"
            )
        else:
            str_output.append(f"{name}:\n{df}\n")
    return "\n".join(str_output)


def format_df_with_info(df_dict: dict[pd.DataFrame]) -> str:
    return format_df_details(df_dict, get_info_dict_from_df_dict(df_dict))


def format_df_with_describe(output_df, name: str = None) -> str:
    if isinstance(output_df, pd.Series):
        output_df = output_df.to_frame()
    if isinstance(output_df, list):
        return "\n".join([format_df_with_describe(df) for df in output_df])
    if isinstance(output_df, dict):
        return "\n".join(
            [format_df_with_describe(df, name) for name, df in output_df.items()]
        )
    if not isinstance(output_df, pd.DataFrame):
        return str(output_df)

    name = name or "Dataframe"
    if output_df.size > 100:
        df_display = pd.concat([output_df.head(50), output_df.tail(50)], axis=0)
        df_string = f"{name} snapshot:\n{_df_to_string(df_display)}\n\nOutput of `df.describe()`:\n{_df_to_string(output_df.describe())}"
    else:
        df_string = f"{name}:\n{_df_to_string(output_df)}"
    return df_string


def _df_to_string(output_df: pd.DataFrame) -> str:
    output_df.columns = [str(col) for col in output_df.columns.tolist()]
    # convert all datetime columns to datetime objects
    datetimecols = [
        col
        for col in output_df.columns.tolist()
        if ("date" in col.lower() or "time" in col.lower())
        and output_df[col].dtype != np.number
    ]
    if "timestamp" in output_df.columns and "timestamp" not in datetimecols:
        datetimecols.append("timestamp")
    for col in datetimecols:
        output_df[col] = output_df[col].astype(dtype="datetime64[ns]", errors="ignore")
        output_df.loc[:, col] = pd.to_datetime(output_df[col], errors="ignore")

    datetimecols = output_df.select_dtypes(include=["datetime64"]).columns.tolist()
    formatters = {col: _format_date for col in datetimecols}
    return output_df.to_string(
        float_format="{:,.2f}".format,
        formatters=formatters,
        na_rep="None",
    )


def _format_date(date: pd.Timestamp):
    return date.strftime("%d %b %Y %H:%M")


def run_n_times(max_tries=1, *, logger: logging.Logger, log_messages={}):
    def dec_wrapper(func):
        def wrapped_func(*args, **kwargs):
            result = None
            if "start" in log_messages:
                logger.info(log_messages["start"])
            for i in range(max_tries):
                logger.info(f"Try num: {i + 1}")
                try:
                    result = func(*args, **kwargs)
                    if result is not None:
                        logger.info(f"Result recieved: {result}")
                        break
                except Exception as e:
                    logger.error(
                        f"Error in try {i + 1}. {e.__class__.__name__}: {e}. Traceback: {e.__traceback__}"
                    )
            if result is None:
                logger.info("Result is None")
            if "end" in log_messages:
                logger.info(log_messages["end"])
            return result

        return wrapped_func

    return dec_wrapper



# standard library imports
import os
import uuid
import logging
from typing import Literal, Union

# third-party imports
import pandas as pd

# local imports
from lyzr.base.errors import (
    MissingValueError,
)
from lyzr.base.file_utils import read_file
from lyzr.data_analyzr.db_connector import (
    DatabaseConnector,
    SQLiteConnector,
)
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore


def get_db_details(
    db_scope: Literal["sql", "ml", "skip"],
    db_type: Literal[
        "files",
        "redshift",
        "postgres",
        "sqlite",
    ],
    config: dict,
    vector_store_config: dict,
    logger: logging.Logger,
):
    df_dict = None
    connector = None
    vector_store = None
    if db_type == "files":
        if "datasets" not in config:
            raise MissingValueError(
                "`datasets` is a required parameter for db_type='files'."
            )
        datasets = config["datasets"]
        files_kwargs = config.get("files_kwargs", {})
        if not isinstance(datasets, dict):
            datasets = {"Dataset": datasets}
        logger.info(
            "Reading the following datasets:\n"
            + "\n".join([f"{df} from {datasets[df]}" for df in datasets])
            + "\n"
        )
        df_dict = get_dict_of_files(datasets, files_kwargs)
        logger.info(
            "Datasets read successfully:\n"
            + "\n".join([f"{df} with shape {df_dict[df].shape}" for df in df_dict])
            + "\n"
        )
    else:
        connector = DatabaseConnector.get_connector(db_type)(**config)

    if db_scope == "ml" and df_dict is None:
        df_dict = connector.fetch_dataframes_dict()
    if df_dict is not None:
        df_keys = list(df_dict.keys())
        for key in df_keys:
            k_new = key.lower().replace(" ", "_")
            df_dict[k_new] = df_dict.pop(key)
    if db_scope == "sql":
        if connector is None:
            connector = SQLiteConnector()
            connector.create_database(
                db_path=config.get("db_path", f"./sqlite/{str(uuid.uuid4())}.db"),
                df_dict=df_dict,
            )
        if isinstance(vector_store_config, dict):
            vector_store = ChromaDBVectorStore(
                path=vector_store_config.get("path", f"./chromadb/{str(uuid.uuid4())}"),
                remake_store=vector_store_config.get(
                    "remake_store", vector_store_config.get("remake", True)
                ),
                connector=connector,
                logger=logger,
            )
    return connector, df_dict, vector_store


def get_dict_of_files(datasets: dict, kwargs) -> dict[pd.DataFrame]:
    # kwargs = {"encoding": "utf-8", "sep": ["\t", ","]}
    kwargs_list = get_list_of_kwargs(datasets, kwargs)
    datasets_dict = {}
    for idx, (name, data) in enumerate(datasets.items()):
        datas = read_file_or_folder(name, data, kwargs_list[idx])
        # print(type(datas), len(datas), datas.keys())
        for d in datas:
            datasets_dict.update({d: datas[d]})
    return datasets_dict


def read_file_or_folder(
    name: str, filepath: Union[str, pd.DataFrame], kwargs
) -> dict[pd.DataFrame]:
    if isinstance(filepath, pd.DataFrame):
        return {name: filepath}
    if os.path.isfile(filepath):
        return {name: read_file(filepath, **kwargs)}
    elif os.path.isdir(filepath):
        return {
            f"{name} {os.path.splitext(file)[0]}": read_file(
                os.path.join(filepath, file), **kwargs
            )
            for file in os.listdir(filepath)
            if os.path.isfile(
                os.path.join(filepath, file)
            )  # do not read from nested directories
        }

    else:
        raise ValueError(
            f"Filepath '{filepath}' is not a valid file or directory. Please provide a valid filepath."
        )


def get_list_of_kwargs(datasets: dict, kwargs: Union[dict, list]) -> list[dict]:
    if isinstance(kwargs, list) and len(kwargs) == len(datasets):
        return kwargs
    kwargs_list = [{} for _ in range(len(datasets))]
    for key, value in kwargs.items():
        if isinstance(value, list) and len(value) == len(
            datasets
        ):  # {"key": [value1, value2, value3]} -> [{"key": value1}, {"key": value2}, {"key": value3}]
            for i, v in enumerate(value):
                kwargs_list[i][key] = v
        else:  # {"key": value} -> [{"key": value}, {"key": value}, {"key": value}]
            for i in range(len(datasets)):
                kwargs_list[i][key] = value
    return kwargs_list



# standard library imports
import os
import json
import uuid
import logging

# local imports
from lyzr.base.errors import (
    ValidationError,
    MissingValueError,
    DependencyError,
)
from lyzr.data_analyzr.db_connector import (
    DatabaseConnector,
    TrainingPlanItem,
    TrainingPlan,
)


class ChromaDBVectorStore:
    def __init__(
        self,
        path: str = None,  # Path to the directory where the ChromaDB data is stored
        remake_store: bool = False,  # If True, the store will be recreated
        connector: DatabaseConnector = None,  # Database connector
        logger: logging.Logger = None,
    ):
        if path is None:
            raise MissingValueError("path")

        if os.path.exists(path):
            self.make_chroma_client(path=path)
            if remake_store:
                logger.info(f"Remaking vector store at {path}.")
                if self.remake_vector_store():
                    logger.info("Generating and adding training plan.\n")
                    self.add_training_plan(plan=connector.get_default_training_plan())
                else:
                    logger.error(f"Failed to recreate vector store at {path}.")
            else:
                logger.info(f"Vector store exists at {path}. Using existing store.")
        else:
            os.makedirs(path)
            logger.info(f"Creating vector store at {path}.")
            self.make_chroma_client(path=path)
            logger.info("Generating and adding training plan.\n")
            self.add_training_plan(plan=connector.get_default_training_plan())

    def make_chroma_client(self, path: str):
        try:
            import chromadb
            from chromadb.config import Settings
            from chromadb.utils import embedding_functions
        except ImportError:
            raise DependencyError(
                {
                    "chromadb": "chromadb==0.4.22",
                }
            )

        self.embedding_function = embedding_functions.DefaultEmbeddingFunction()

        self.chroma_client = chromadb.PersistentClient(
            path=path, settings=Settings(anonymized_telemetry=False)
        )
        self.sql_collection = self.chroma_client.get_or_create_collection(
            name="sql", embedding_function=self.embedding_function
        )
        self.documentation_collection = self.chroma_client.get_or_create_collection(
            name="documentation", embedding_function=self.embedding_function
        )
        self.ddl_collection = self.chroma_client.get_or_create_collection(
            name="ddl", embedding_function=self.embedding_function
        )

    def remake_vector_store(self):
        remake = True
        for col in self.chroma_client.list_collections():
            remake = remake and self.remake_collection(collection_name=col.name)
        return remake

    def add_training_plan(
        self,
        question: str = None,
        sql: str = None,
        ddl: str = None,
        documentation: str = None,
        plan: TrainingPlan = None,
    ) -> str:
        if question and not sql:
            raise ValidationError("Please also provide a SQL query")

        if documentation:
            return self.add_documentation(documentation)

        if sql:
            return self.add_question_sql(question=question, sql=sql)

        if ddl:
            return self.add_ddl(ddl)

        if plan:
            for item in plan._plan:
                if item.item_type == TrainingPlanItem.ITEM_TYPE_DDL:
                    self.add_ddl(item.item_value)
                elif item.item_type == TrainingPlanItem.ITEM_TYPE_IS:
                    self.add_documentation(item.item_value)
                elif item.item_type == TrainingPlanItem.ITEM_TYPE_SQL:
                    self.add_question_sql(question=item.item_name, sql=item.item_value)

    def remake_collection(self, collection_name: str) -> bool:
        if collection_name == "sql":
            self.chroma_client.delete_collection(name="sql")
            self.sql_collection = self.chroma_client.get_or_create_collection(
                name="sql", embedding_function=self.embedding_function
            )
            return True
        elif collection_name == "ddl":
            self.chroma_client.delete_collection(name="ddl")
            self.ddl_collection = self.chroma_client.get_or_create_collection(
                name="ddl", embedding_function=self.embedding_function
            )
            return True
        elif collection_name == "documentation":
            self.chroma_client.delete_collection(name="documentation")
            self.documentation_collection = self.chroma_client.get_or_create_collection(
                name="documentation", embedding_function=self.embedding_function
            )
            return True
        else:
            return False

    def _extract_documents(self, query_results) -> list:
        if query_results is None:
            return []

        if "documents" in query_results:
            documents = query_results["documents"]
            if len(documents) == 1 and isinstance(documents[0], list):
                try:
                    documents = [json.loads(doc) for doc in documents[0]]
                except Exception:
                    return documents[0]

            return documents

    def get_related_ddl(self, question: str) -> list:
        return self._extract_documents(
            self.ddl_collection.query(
                query_texts=[question],
            )
        )

    def get_related_documentation(self, question: str) -> list:
        return self._extract_documents(
            self.documentation_collection.query(
                query_texts=[question],
            )
        )

    def get_similar_question_sql(self, user_input: str) -> list:
        return self._extract_documents(
            self.sql_collection.query(
                query_texts=[user_input],
            )
        )

    def generate_embedding(self, data: str) -> list[float]:
        embedding = self.embedding_function([data])
        if len(embedding) == 1:
            return embedding[0]
        return embedding

    def add_question_sql(self, question: str, sql: str) -> str:
        question_sql_json = json.dumps(
            {
                "question": question,
                "sql": sql,
            }
        )
        sql_id = str(uuid.uuid4()) + "-sql"
        self.sql_collection.add(
            documents=question_sql_json,
            embeddings=self.generate_embedding(question_sql_json),
            ids=sql_id,
        )
        return sql_id

    def add_ddl(self, ddl: str) -> str:
        ddl_id = str(uuid.uuid4()) + "-ddl"
        self.ddl_collection.add(
            documents=ddl,
            embeddings=self.generate_embedding(ddl),
            ids=ddl_id,
        )
        return ddl_id

    def add_documentation(self, documentation: str) -> str:
        doc_id = str(uuid.uuid4()) + "-doc"
        self.documentation_collection.add(
            documents=documentation,
            embeddings=self.generate_embedding(documentation),
            ids=doc_id,
        )
        return doc_id



# standard library imports
import os
import time
import uuid
import warnings
from typing import Union, Literal, Optional, Any

# third-party imports
import numpy as np
import pandas as pd

# local imports
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.data_analyzr.file_utils import get_db_details
from lyzr.data_analyzr.txt_to_sql_utils import TxttoSQLFactory
from lyzr.data_analyzr.ml_analysis_utils import MLAnalysisFactory
from lyzr.base.errors import (
    MissingValueError,
)
from lyzr.base.llm import (
    LyzrLLMFactory,
    LiteLLM,
)
from lyzr.data_analyzr.plot_utils import PlotFactory
from lyzr.data_analyzr.utils import (
    format_df_with_describe,
    format_df_with_info,
    get_info_dict_from_df_dict,
)

# imports for logging
import sys
import logging

# imports for legacy usage
from PIL import Image
from pathlib import Path
from pandas.errors import EmptyDataError

warnings.filterwarnings("ignore", category=UserWarning, module="pydantic")


class DataAnalyzr:

    def __init__(
        self,
        analysis_type: Optional[Literal["sql", "ml", "skip"]] = None,
        api_key: Optional[str] = None,
        max_retries: Optional[int] = 5,
        model: Optional[LiteLLM] = None,
        user_input: Optional[str] = None,
        context: Optional[str] = None,
        log_level: Optional[
            Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        ] = "INFO",
        print_log: Optional[bool] = False,
        log_filename: Optional[str] = "dataanalyzr.log",
        # legacy usage
        df: Union[str, pd.DataFrame] = None,
        model_type: Optional[str] = None,
        model_name: Optional[str] = None,
        seed: Optional[int] = 0,
    ):
        api_key = (
            api_key or os.environ.get("API_KEY") or os.environ.get("OPENAI_API_KEY")
        )
        if api_key is None:
            raise MissingValueError(
                "Provide a value for `api_key` or set the `OPENAI_API_KEY` environment variable."
            )
        self.max_retries = max_retries
        self.user_input = user_input
        self.log_filename = log_filename
        self._set_logger(log_level, print_log)

        if df is not None:
            # legacy usage
            self._legacy_usage(
                api_key=api_key,
                df=df,
                user_input=user_input,
                model=model,
                model_type=model_type,
                model_name=model_name,
                seed=seed,
            )
        else:
            if analysis_type is None:
                raise MissingValueError("`analysis_type` is a required parameter.")
            if model is None:
                self.model = LyzrLLMFactory().from_defaults(
                    model="gpt-4-1106-preview", api_key=api_key, seed=seed
                )
            elif isinstance(model, LiteLLM):
                self.model = model
            self.model.additional_kwargs["logger"] = self.logger

            self.analysis_type = analysis_type.lower().strip()
            self.context = context or ""
        (
            self.dataset_description_output,
            self.ai_queries,
            self.analysis_output,
            self.visualisation_output,
            self.insights_output,
            self.recommendations_output,
            self.tasks_output,
        ) = (None, None, None, None, None, None, None)

    def _legacy_usage(
        self,
        api_key: str,
        df: Union[str, pd.DataFrame],
        user_input: Optional[str],
        model: Optional[LiteLLM],
        model_type: Optional[str],
        model_name: Optional[str],
        seed: int,
    ):
        warnings.warn(
            "The `df` parameter is deprecated and will be removed in a future version. Please use the `get_data` method to load data."
        )
        for param in ["model", "model_type", "model_name", "seed"]:
            if locals()[param] is not None:
                warnings.warn(
                    f"The `{param}` parameter is deprecated and will be removed in a future version. Please use the `analysis_model` parameter to set the analysis model, and the `gen_model` parameter to set the generation model."
                )
        self.model = model or LyzrLLMFactory().from_defaults(
            api_key=api_key,
            api_type=model_type,
            model=model_name or os.environ.get("MODEL_NAME", "gpt-4-1106-preview"),
            seed=seed,
        )
        self.context = ""
        self.analysis_type = "ml"
        self.user_input = user_input

        def _clean_df(df: pd.DataFrame):
            df = df[df.columns[df.isnull().mean() < 0.5]]
            cat_columns = df.select_dtypes(include=["object"]).columns
            num_columns = df.select_dtypes(include=[np.number]).columns
            df[cat_columns] = df[cat_columns].apply(lambda x: x.fillna(x.mode()[0]))
            df[num_columns] = df[num_columns].apply(lambda x: x.fillna(x.mean()))
            df = df.loc[:, ~df.columns.str.contains("^Unnamed")]
            df = df.drop_duplicates(keep="first")
            return df

        if isinstance(df, str):
            self.database_connector, self.df_dict, self.vector_store = get_db_details(
                "ml", "files", {"datasets": {"Dataset": df}}, {}, logger=self.logger
            )
            for name, df in self.df_dict.items():
                if not isinstance(df, pd.DataFrame):
                    continue
                self.df_dict[name] = _clean_df(df)
        elif isinstance(df, pd.DataFrame):
            if df.empty:
                raise EmptyDataError("The provided DataFrame is empty.")
            self.df_dict = {"dataset": _clean_df(df)}
            self.database_connector = None
            self.vector_store = None
        else:
            raise ValueError("df must be a path to a file or a pd.DataFrame object.")
        self.df_info_dict = get_info_dict_from_df_dict(self.df_dict)

    def _set_logger(self, log_level, print_log):
        self.logger = logging.getLogger(__name__)
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError("Invalid log level: %s" % log_level)
        self.logger.setLevel(numeric_level)

        if self.logger.hasHandlers():
            for handler in self.logger.handlers:
                try:
                    handler.close()
                except Exception:
                    pass
            self.logger.handlers.clear()

        if print_log:
            # output logs to stdout
            handler = logging.StreamHandler(sys.stdout)
            handler.setLevel(numeric_level)
            self.logger.addHandler(handler)

        log_filename = self.log_filename
        dir_path = os.path.dirname(log_filename)
        if dir_path.strip() != "":
            os.makedirs(dir_path, exist_ok=True)
        file_handler = logging.FileHandler(
            log_filename, mode="a"
        )  # Open the log file in append mode
        file_handler.setLevel(numeric_level)

        # Optionally, you can set a formatter for the file handler if you want a different format for file logs
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s\n%(message)s\n",
            datefmt="%d-%b-%y %H:%M:%S",
        )
        formatter.converter = time.gmtime
        file_handler.setFormatter(formatter)

        # Add the file handler to the logger
        self.logger.addHandler(file_handler)

    # Function to get datasets
    def get_data(
        self,
        db_type: Literal[
            "files",
            "redshift",
            "postgres",
            "sqlite",
        ],
        config: dict,
        # if db_type == "files", config_keys = datasets, files_kwargs, db_path
        # if db_type == "redshift" or "postgres" or "snowflake" or "mysql", config_keys = host, port, user, password, database, schema, table
        # if db_type == "snowflake", config_keys = warehouse, role
        # if db_type == "sqlite", config_keys = db_path
        vector_store_config,
    ) -> Any:
        self.database_connector, self.df_dict, self.vector_store = get_db_details(
            self.analysis_type, db_type, config, vector_store_config, self.logger
        )
        self.df_info_dict = (
            get_info_dict_from_df_dict(self.df_dict)
            if self.df_dict is not None
            else None
        )

    def analysis(
        self,
        user_input: str,
        analysis_context: str,
        analysis_steps: dict = None,
    ):
        if self.analysis_type == "skip" and analysis_steps is None:
            if self.df_dict is None:
                self.logger.info(
                    "No analysis performed. Fetching dataframes from database."
                )
                self.df_dict = self.database_connector.fetch_dataframes_dict()
            self.analysis_output = self.df_dict
            self.analysis_guide = (
                "No analysis performed. Analysis output is the given dataframe."
            )
            return self.analysis_output
        analysis_model = LyzrLLMFactory().from_defaults(model="gpt-3.5-turbo")
        analysis_model.additional_kwargs["logger"] = self.logger
        if self.analysis_type == "sql" and analysis_steps is None:
            return self._txt_to_sql_analysis(
                analysis_model, user_input, analysis_context
            )
        if self.analysis_type == "ml" or analysis_steps is not None:
            return self._ml_analysis(
                analysis_model,
                user_input,
                analysis_context,
                analysis_steps=analysis_steps,
            )

    def _ml_analysis(
        self,
        analysis_model,
        user_input: str,
        analysis_context: str = None,
        analysis_steps: dict = None,
    ):
        self.analyzer = MLAnalysisFactory(
            model=analysis_model,
            data_dict=self.df_dict,
            data_info_dict=self.df_info_dict,
            logger=self.logger,
            context=analysis_context,
        )
        if analysis_steps is not None:
            _, data = self.analyzer.get_analysis_from_steps(analysis_steps)
            return data
        else:
            self.analysis_output = self.analyzer.run_complete_analysis(user_input)
            self.analysis_guide = self.analyzer.analysis_guide
            return self.analysis_output

    def _txt_to_sql_analysis(
        self, analysis_model, user_input: str, analysis_context: str = None
    ):
        self.analyzer = TxttoSQLFactory(
            model=self.model,
            db_connector=self.database_connector,
            logger=self.logger,
            context=analysis_context,
            vector_store=self.vector_store,
        )
        self.analysis_output = self.analyzer.run_complete_analysis(user_input)
        self.analysis_guide = self.analyzer.analysis_guide
        return self.analysis_output

    def visualisation(
        self,
        user_input: str,
        plot_context: str = None,
        plot_path: str = None,
    ):
        if plot_path is None:
            plot_path = Path(f"generated_plots/{str(uuid.uuid4())}.png").as_posix()
        else:
            plot_path = Path(plot_path).as_posix()
        self.visualisation_output = None
        # self.start_time = time.time()

        self.logger.info("Generating visualisation\n")
        plotter = PlotFactory(
            model=self.model,
            logger=self.logger,
            plot_context=plot_context,
            plot_path=plot_path,
            df_dict=self.df_dict,
            database_connector=self.database_connector,
            analysis_type=self.analysis_type,
            analyzer=self.analyzer,
            analysis_output=self.analysis_output,
        )
        self.visualisation_output = plotter.get_visualisation(user_input)
        return self.visualisation_output
        # while True:
        #     try:
        #         self.logger.info("Generating visualisation\n")
        #         plotter = PlotFactory(
        #             model=self.model,
        #             plotting_model_kwargs=self._plot_model_kwargs,
        #             df_dict=plot_df,
        #             logger=self.logger,
        #             plot_context=plot_context,
        #             plot_path=plot_path,
        #             use_guide=use_guide,
        #         )
        #         analysis_steps = plotter.get_plotting_and_analysis_steps(
        #             self.user_input
        #         )
        #         if analysis_steps is not None and "steps" in analysis_steps:
        #             if len(analysis_steps["steps"]) == 0:
        #                 self.plot_df = self.df_dict[analysis_steps["df_name"]]
        #             else:
        #                 self.plot_df = self.analysis(user_input, "", analysis_steps)
        #         elif not use_guide:
        #             self.plot_df = plot_df[list(plot_df.keys())[0]]
        #         else:
        #             self.logger.info(
        #                 "No analysis steps found. Using first dataframe for plotting.\n"
        #             )
        #             self.plot_df = self.df_dict[list(self.df_dict.keys())[0]]

        #         self.visualisation_output = plotter.get_visualisation(self.plot_df)
        #         return self.visualisation_output
        #     except RecursionError:
        #         plt.close()
        #         raise RecursionError(
        #             "The request could not be completed. Please wait a while and try again."
        #         )
        #     except Exception as e:
        #         plt.close()
        #         if time.time() - self.start_time > 30:
        #             raise TimeoutError(
        #                 "The request could not be completed. Please wait a while and try again."
        #             )
        #         self.logger.info(f"{e.__class__.__name__}: {e}")
        #         self.logger.info("Traceback:\n{}\n".format(traceback.format_exc()))
        #         continue

    def insights(
        self,
        user_input: str,
        insights_context: Optional[str] = None,
        n_insights: Optional[int] = 3,
    ) -> str:
        if "analysis_guide" not in self.__dict__:
            self.analysis_guide = ""
        self.insights_output = self.model.run(
            messages=[
                LyzrPromptFactory(name="insights", prompt_type="system").get_message(
                    context=insights_context.strip() + "\n\n", n_insights=n_insights
                ),
                LyzrPromptFactory(name="insights", prompt_type="user").get_message(
                    user_input=user_input,
                    analysis_guide=self.analysis_guide,
                    analysis_output=format_df_with_describe(self.analysis_output),
                    date=time.strftime("%d %b %Y"),
                ),
            ],
            temperature=0.3,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0,
        ).message.content.strip()
        return self.insights_output

    def recommendations(
        self,
        user_input: Optional[str] = None,
        use_insights: Optional[bool] = True,
        recs_format: Optional[dict] = None,
        recommendations_context: Optional[str] = None,
        n_recommendations: Optional[int] = 3,
        output_type: Optional[Literal["text", "json"]] = "text",
        # legacy usage
        insights: Optional[str] = None,
        schema: Optional[list] = None,
    ) -> str:
        if insights is not None or schema is not None:
            use_insights = True
            output_type = "json"
        if insights is not None:
            warnings.warn(
                "The `insights` parameter is deprecated and will be removed in a future version. Please use the `insights_output` attribute to set the insights."
            )
            self.insights_output = insights
        if schema is not None:
            warnings.warn(
                "The `schema` parameter is deprecated and will be removed in a future version. Please use the `recs_format` parameter to set the recommendations format as a dictionary."
            )
            for elem in schema:
                if isinstance(elem, dict):
                    recs_format = elem
                    break

        system_message_sections = ["context"]
        system_message_dict = {}
        user_message_dict = {
            "user_input": user_input,
            "analysis_output": f"Analysis output:\n{format_df_with_describe(self.analysis_output)}",
        }
        if recommendations_context is not None and recommendations_context != "":
            system_message_sections.append("external_context")
            system_message_dict["context"] = recommendations_context.strip() + "\n\n"

        if not use_insights:
            insights = None
            system_message_sections.append("task_no_insights")
        else:
            system_message_sections.append("task_with_insights")
            user_message_dict["insights"] = (
                f"\nAnalysis insights:\n{self.insights_output}\n"
            )

        if output_type.lower().strip() == "json":
            recs_format = recs_format or {
                "Recommendation": "string",
                "Basis of the Recommendation": "string",
                "Impact if implemented": "string",
            }
            system_message_sections.append("json_type")
            system_message_dict["output_json_keys"] = ", ".join(recs_format.keys())
            system_message_dict["json_schema"] = [
                recs_format for _ in range(n_recommendations)
            ]
        elif output_type.lower().strip() == "text":
            system_message_sections.append("text_type")
            system_message_dict["n_recommendations"] = n_recommendations

        system_message_sections.append("closing")
        self.recommendations_output = self.model.run(
            messages=[
                LyzrPromptFactory(
                    name="recommendations", prompt_type="system"
                ).get_message(
                    use_sections=system_message_sections, **system_message_dict
                ),
                LyzrPromptFactory(
                    name="recommendations", prompt_type="user"
                ).get_message(**user_message_dict),
            ],
            temperature=0.3,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0,
            response_format=(
                {"type": "json_object"} if output_type == "json" else {"type": "text"}
            ),
        ).message.content
        return self.recommendations_output

    def tasks(
        self,
        user_input: Optional[str] = None,
        tasks_context: Optional[str] = None,
        n_tasks: Optional[int] = 5,
        # legacy usage
        insights: Optional[str] = None,
        recommendations: Optional[str] = None,
    ) -> str:
        if insights is not None:
            warnings.warn(
                "The `insights` parameter is deprecated and will be removed in a future version. Please use the `insights_output` attribute to set the insights."
            )
            self.insights_output = insights
        if recommendations is not None:
            warnings.warn(
                "The `recommendations` parameter is deprecated and will be removed in a future version. Please use the `recommendations_output` attribute to set the recommendations."
            )
            self.recommendations_output = recommendations
        self.tasks_output = self.model.run(
            messages=[
                LyzrPromptFactory(name="tasks", prompt_type="system").get_message(
                    context=tasks_context.strip() + "\n\n", n_tasks=n_tasks
                ),
                LyzrPromptFactory(name="tasks", prompt_type="user").get_message(
                    user_input=user_input or self.user_input,
                    insights=self.insights_output,
                    recommendations=self.recommendations_output,
                ),
            ],
            temperature=1,
            top_p=0.3,
            frequency_penalty=0.7,
            presence_penalty=0.3,
        ).message.content.strip()
        return self.tasks_output

    def ask(
        self,
        user_input: str = None,
        outputs: list[
            Literal["visualisation", "insights", "recommendations", "tasks"]
        ] = None,
        plot_path: str = None,
        rerun_analysis: bool = True,
        use_insights: bool = True,
        recs_format: dict = None,
        recs_output_type: Literal["text", "json"] = None,
        counts: dict = None,  # keys: insights, recommendations, tasks
        context: dict = None,  # keys: analysis, insights, recommendations, tasks
    ):
        if use_insights is None:
            use_insights = True
        if rerun_analysis is None:
            rerun_analysis = True
        if outputs is None:
            outputs = ["visualisation", "insights", "recommendations", "tasks"]
        if user_input is None and self.user_input is None:
            raise MissingValueError(
                "`user_input` is a required parameter to generate outputs."
            )
        if user_input is None:
            user_input = self.user_input
        context = context or {}
        counts = counts or {}

        if self.analysis_output is None or rerun_analysis:
            self.analysis_output = self.analysis(
                user_input=user_input,
                analysis_context=context.get("analysis", self.context),
            )

        if "visualisation" in outputs:
            self.visualisation_output = self.visualisation(
                user_input=user_input,
                plot_context=context.get("visualisation", self.context),
                plot_path=plot_path,
            )
        else:
            self.visualisation_output = ""

        if (
            "insights" in outputs
            or "tasks" in outputs
            or ("recommendations" in outputs and use_insights)
        ):
            self.insights_output = self.insights(
                user_input=user_input,
                insights_context=context.get("insights", self.context),
                n_insights=counts.get("insights", 3),
            )
        else:
            self.insights_output = ""

        if "recommendations" in outputs or "tasks" in outputs:
            self.recommendations_output = self.recommendations(
                user_input=user_input,
                use_insights=use_insights,
                recs_format=recs_format,
                recommendations_context=context.get("recommendations", self.context),
                n_recommendations=counts.get("recommendations", 3),
                output_type=recs_output_type or "text",
            )
        else:
            self.recommendations_output = ""

        if "tasks" in outputs:
            self.tasks_output = self.tasks(
                user_input=user_input,
                tasks_context=context.get("tasks", self.context),
                n_tasks=counts.get("tasks", 5),
            )
        else:
            self.tasks_output = ""

        return {
            "visualisation": self.visualisation_output,
            "insights": self.insights_output,
            "recommendations": self.recommendations_output,
            "tasks": self.tasks_output,
        }

    # ---------------------------------------- Legacy functions, for backward compatibility ----------------------------------------
    # ------------------------------ These functions are not used in the current version of the code. ------------------------------

    def analysis_insights(self, user_input: str) -> str:
        warnings.warn(
            "The `analysis_insights` method is deprecated and will be removed in a future version. Use the `insights` method instead."
        )
        self.analysis_output = self.analysis(
            user_input=user_input,
            analysis_context="",
        )
        return self.insights(user_input)

    def visualizations(
        self, user_input: str, dir_path: Path = None
    ) -> list[Image.Image]:
        warnings.warn(
            "The `visualizations` method is deprecated and will be removed in a future version. Use the `visualisation` method instead."
        )
        return [
            Image.open(
                self.visualisation(
                    user_input,
                    plot_path=(
                        str(dir_path / "plot.png") if dir_path is not None else None
                    ),
                )
            )
        ]

    def dataset_description(self, context: str = None) -> str:
        warnings.warn(
            "The `dataset_description` method is deprecated and will be removed in a future version."
        )
        context = context or self.context
        df_desc_dict = {}
        for name, df in self.df_dict.items():
            df_desc_dict[name] = self.model.run(
                messages=[
                    LyzrPromptFactory(
                        name="dataset_description", prompt_type="system"
                    ).get_message(context=context.strip() + "\n\n"),
                    LyzrPromptFactory(
                        name="dataset_description", prompt_type="user"
                    ).get_message(
                        df_head=df.head(),
                        headers=df.columns.tolist(),
                    ),
                ],
                temperature=1,
                top_p=0.3,
                frequency_penalty=0.7,
                presence_penalty=0.3,
            ).message.content.strip()
        self.dataset_description_output = "\n".join(
            [f"{name}:\n{desc}" for name, desc in df_desc_dict.items()]
        )
        return self.dataset_description_output

    def ai_queries_df(
        self, dataset_description: Optional[str] = None, context: Optional[str] = None
    ) -> str:
        warnings.warn(
            "The `ai_queries_df` method is deprecated and will be removed in a future version."
        )
        context = context or self.context
        self.dataset_description_output = (
            dataset_description or self.dataset_description_output
        )
        if self.dataset_description_output is None:
            self.dataset_description_output = self.dataset_description(context)

        self.ai_queries = self.model.run(
            messages=[
                LyzrPromptFactory(name="ai_queries", prompt_type="system").get_message(
                    context=context.strip() + "\n\n"
                ),
                LyzrPromptFactory(name="ai_queries", prompt_type="user").get_message(
                    df_details=format_df_with_describe(self.df_dict)
                ),
            ],
            temperature=1,
            top_p=0.3,
            frequency_penalty=0.7,
            presence_penalty=0.3,
        ).message.content.strip()

        return self.ai_queries

    def analysis_recommendation(
        self,
        user_input: Optional[str] = None,
        number_of_recommendations: Optional[int] = 4,
    ):
        warnings.warn(
            "The `analysis_recommendation` method is deprecated and will be removed in a future version."
        )
        formatted_user_input: str = (
            LyzrPromptFactory("format_user_input", "system")
            .get_message(user_input=user_input)
            .content
            if user_input is not None
            else ""
        )
        recommendations = self.model.run(
            messages=[
                LyzrPromptFactory("analysis_recommendations", "system").get_message(
                    number_of_recommendations=number_of_recommendations,
                    df_details=format_df_with_info(self.df_dict),
                    formatted_user_input=formatted_user_input,
                ),
            ],
            temperature=0.2,
        ).message.content.strip()
        return recommendations



import re
import logging
from typing import Literal
from itertools import chain

all_tasks = {
    "clean_data": [
        "remove_nulls",
        "convert_to_datetime",
        "convert_to_numeric",
        "convert_to_categorical",
    ],
    "transform": [
        "one_hot_encode",
        "ordinal_encode",
        "scale",
        "extract_time_period",
        "select_indices",
    ],
    "math_operation": ["add", "subtract", "multiply", "divide"],
    "analysis": [
        "sortvalues",
        "filter",
        "mean",
        "sum",
        "cumsum",
        "groupby",
        "correlation",
        "regression",
        "classification",
        "cluster",
        "forecast",
    ],
}

# ------------------------------ LLM output extraction --------------------------------------


def extract_sql(llm_response: str, logger: logging.Logger) -> str:
    # If the llm_response contains a markdown code block, with or without the sql tag, extract the sql from it
    sql = re.search(r"```sql\n(.*)```", llm_response, re.DOTALL)
    if sql:
        logger.info(f"Output from LLM: {llm_response} \nExtracted SQL: {sql.group(1)}")
        return sql.group(1)

    sql = re.search(r"```(.*)```", llm_response, re.DOTALL)
    if sql:
        logger.info(f"Output from LLM: {llm_response} \nExtracted SQL: {sql.group(1)}")
        return sql.group(1)

    return llm_response


# ------------------------------ LLM output validation --------------------------------------


def validate_output_step_details(step_details: dict, logger: logging.Logger):
    task_details, step_details = _check_task_details(step_details)
    if not task_details:
        logger.error("Invalid task details: {}.".format(step_details))
        raise ValueError("Invalid task details: {}".format(step_details))
    if not _check_task_and_type(step_details):
        logger.error(
            "Invalid task: {}. Invalid task details: {}.".format(
                step_details["task"], step_details
            )
        )
        raise ValueError(
            "Invalid task: {}. Invalid task details: {}.".format(
                step_details["task"], step_details
            )
        )
    return step_details


def _get_task_from_type(task: str, task_type: str) -> tuple:
    all_task_types = list(chain.from_iterable(all_tasks.values()))
    if (task_type not in all_task_types) and (task in all_task_types):
        task_type = task
    for task, types in all_tasks.items():
        if task_type in types:
            return task, task_type


def _check_task_details(step_details: dict) -> tuple:
    if (
        "task" in step_details
        and "type" in step_details
        and "args" in step_details
        and isinstance(step_details["args"], dict)
    ):
        return True, step_details
    if "task" not in step_details and "type" in step_details:
        step_details["task"] = ""
        return True, step_details
    if "type" not in step_details and "task" in step_details:
        if step_details["task"] in list(chain.from_iterable(all_tasks.values())):
            step_details["type"] = step_details["task"]
            return True, step_details
    if "args" not in step_details or not isinstance(step_details["args"], dict):
        step_details["args"] = {}
        return True, step_details
    return False, step_details


def _check_task_and_type(step_details: dict) -> bool:
    if (step_details["task"] not in all_tasks.keys()) or (
        step_details["type"] not in all_tasks[step_details["task"]]
    ):
        step_details["task"], step_details["type"] = _get_task_from_type(
            step_details["task"], step_details["type"]
        )
    if step_details["task"] not in all_tasks.keys():
        return False
    return True


# ---------------------------------- LLM output format checking ------------------------------------------------


# for output format dict
def check_output_format(
    llm_output: str,
    logger: logging.Logger,
    output_type: Literal["plot", "analysis"] = "analysis",
) -> dict:
    try:
        llm_output = eval(llm_output)
        logger.info("LLM output format checked.\n")
        return llm_output
    except NameError as e:
        e = str(e)
        if e.startswith("name") and e.endswith("is not defined"):
            word = re.findall(r"['\"](.*)['\"]", e)[0]
            llm_output = _check_word(llm_output, word)
            return check_output_format(llm_output, logger)
    except Exception as e:
        logger.info("Invalid output from LLM. Attempting to fix.")

    if output_type == "analysis":
        # remove all spaces in string
        llm_output = re.sub(r"\s+", "", llm_output)
        # string checking
        n_steps = _get_number_of_steps(llm_output)
        steps_components = _get_component_elements(llm_output)
        error_components = _check_num_components(steps_components, n_steps)
        if len(error_components) != 0:
            logger.info(f"Fixing components: {', '.join(error_components)}.")
            llm_output = _fix_components(error_components, steps_components)

        return check_output_format(llm_output, logger)


def _get_component_elements(llm_output):
    # string matching
    quotes = r"('|\")"
    number_regex = r"\d+"
    word_regex = rf"{quotes}\w+{quotes}"
    alphanum_regex = rf"({word_regex}|{number_regex})"
    list_0_regex = rf"\[{alphanum_regex}(,{alphanum_regex})*?\]"
    tasks_regex = r"['\"]task['\"]:(.*?),"
    task_types_regex = r"['\"]type['\"]:(.*?),"
    output_columns_regex = (r"['\"]outputcolumns['\"]:(.*)", r"['\"](\w+)['\"]")
    output_columns_list_regex = rf"{quotes}outputcolumns{quotes}:(\[{word_regex}(,{word_regex})+?\])"  # list of alphanumeric strings
    args_dict_regex = (
        "{"
        + f"{word_regex}:({word_regex}|{number_regex}|{list_0_regex})(,{word_regex}:({word_regex}|{number_regex}|{list_0_regex}))*?"
        + "}"  # dict with string keys and alphanum or list_0 values
    )
    step_details_dict_regex = (
        "{"
        + f"{word_regex}:({word_regex}|{number_regex}|{list_0_regex}|{args_dict_regex})(,{word_regex}:({word_regex}|{number_regex}|{list_0_regex}|{args_dict_regex}))*?"
        + "}"
    )  # dict with string keys and alphanum or list_0 or args_dict values
    all_steps_regex = rf"\[{step_details_dict_regex}(,{step_details_dict_regex})*?\]"  # list of step_details_dict
    response_dict_regex = (
        "{"
        + f"'steps':{all_steps_regex},'outputcolumns':{output_columns_list_regex}"
        + "}"
    )  # dict with keys 'steps' and 'output columns', and values all_steps and output_columns_list respectively

    steps_components = {
        "tasks": _get_tasks(llm_output, tasks_regex),
        "task_types": _get_task_types(llm_output, task_types_regex),
        "output_columns": _get_output_columns(llm_output, output_columns_regex),
        "output_columns_list": _get_output_columns_list(
            llm_output, output_columns_list_regex
        ),
        "args_dict": _get_args_dicts(llm_output, args_dict_regex),
        "step_details_dict": _get_steps_dicts(llm_output, step_details_dict_regex),
        "all_steps": _get_all_steps(llm_output, all_steps_regex),
        "response_dict": _get_response_dict(llm_output, response_dict_regex),
    }
    return steps_components


def _get_number_of_steps(llm_output):
    number = r"(\w+)"
    num_steps = []
    for word in re.finditer(number, llm_output):
        try:
            num_steps.append(int(word.group()))
        except ValueError:
            llm_output = _check_word(llm_output, word)
    while len(num_steps) != 0:
        if len(num_steps) != max(num_steps):
            # remove highest number and try again
            num_steps.remove(max(num_steps))
        else:
            return max(num_steps)
    return 0


def _check_word(llm_output, word):
    if isinstance(word, str):
        nmatches = len([x for x in re.finditer(word, llm_output)])
        for idx in range(nmatches):
            match = [x for x in re.finditer(word, llm_output)][idx]
            llm_output = _check_word_match(llm_output, match)
        return llm_output
    elif isinstance(word, re.Match):
        return _check_word_match(llm_output, word)


def _check_word_match(llm_output, word):
    start_quote = llm_output[word.start() - 1]
    end_quote = llm_output[word.end()]
    if start_quote not in ["'", '"']:
        if end_quote not in ["'", '"']:
            llm_output = llm_output[: word.start()] + "'" + llm_output[word.start() :]
            llm_output = (
                llm_output[: word.end() + 1] + "'" + llm_output[word.end() + 1 :]
            )
        else:
            llm_output = (
                llm_output[: word.start()] + end_quote + llm_output[word.start() :]
            )
    elif end_quote not in ["'", '"']:
        llm_output = llm_output[: word.end()] + start_quote + llm_output[word.end() :]
    return llm_output


def _check_num_components(steps_components, n_steps):
    error_components = []
    for component in steps_components:
        if component == "output_columns":
            continue
        if component in ["tasks", "task_types", "args_dict", "step_details_dict"]:
            expected_num_components = n_steps
        else:
            expected_num_components = 1
        if len(steps_components[component]) != expected_num_components:
            error_components.append(component)
    return error_components


def _fix_components(error_components, steps_components):
    for component in error_components:
        if component in ["tasks", "task_types", "args_dict", "output_columns"]:
            raise ValueError("Cannot fix LLM output.")
        steps_components[component] = globals()[f"_fix_{component}"](steps_components)
    return steps_components["response_dict"][0]


def _fix_step_details_dict(steps_components):
    step_details = []
    for step, args_dict in enumerate(steps_components["args_dict"]):
        step_details.append(
            "{"
            + f"'step':{step},'task':{steps_components['tasks'][step]},'type':{steps_components['type'][step]},'args':{args_dict}"
            + "}"
        )


def _fix_output_columns_list(steps_components):
    return ["[" + ",".join(steps_components["output_columns"]) + "]"]


def _fix_all_steps(steps_components):
    return ["[" + ",".join(steps_components["step_details_dict"]) + "]"]


def _fix_response_dict(steps_components):
    return [
        "{"
        + f"'steps':{steps_components['all_steps'][0]},'output columns':{steps_components['output_columns_list'][0]}"
        + "}"
    ]


def _get_tasks(llm_output, tasks_regex):
    tasks = []
    for x in re.finditer(tasks_regex, llm_output):
        tasks.append(x.group(1))
    return tasks


def _get_task_types(llm_output, task_types_regex):
    task_types = []
    for x in re.finditer(task_types_regex, llm_output):
        task_types.append(x.group(1))
    return task_types


def _get_output_columns(llm_output, output_columns_regex):
    output_columns = []
    for x in re.finditer(output_columns_regex[0], llm_output):
        for y in re.finditer(output_columns_regex[1], x.group(1)):
            output_columns.append(y.group(1))
    return output_columns


def _get_output_columns_list(llm_output, output_columns_list_regex):
    output_columns_list = []
    for x in re.finditer(output_columns_list_regex, llm_output):
        output_columns_list.append(x.group(3))
    return output_columns_list


def _get_args_dicts(llm_output, args_dict_regex):
    args_dicts = []
    for x in re.finditer(args_dict_regex, llm_output):
        args_dicts.append(x.group())
    return args_dicts


def _get_steps_dicts(llm_outptut, steps_dicts_regex):
    steps_dicts = []
    for x in re.finditer(steps_dicts_regex, llm_outptut):
        steps_dicts.append(x.group())
    return steps_dicts


def _get_all_steps(llm_output, all_steps_regex):
    all_steps = []
    for x in re.finditer(all_steps_regex, llm_output):
        all_steps.append(x.group())
    return all_steps


def _get_response_dict(llm_output, response_dict_regex):
    response_dict = []
    for x in re.finditer(response_dict_regex, llm_output):
        response_dict.append(x.group())
    return response_dict



# standard library imports
import re
import time
import logging
import traceback
from typing import Union, Optional, Literal, Any

# third-party imports
import pandas as pd

# local imports
from lyzr.base.base import SystemMessage, AssistantMessage
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.llm import LiteLLM
from lyzr.data_analyzr.output_handler import (
    check_output_format,
    validate_output_step_details,
)
from lyzr.base.errors import DependencyError
from lyzr.data_analyzr.utils import get_columns_names


def run_analysis_step(
    data: pd.DataFrame,
    step_details: dict,
    logger: logging.Logger,
):
    logger.info(f"\nRunning step: {step_details}")
    if step_details["task"] == "clean_data":
        step_details["args"]["columns"] = get_columns_names(
            df_columns=data.columns,
            arguments=step_details["args"],
            logger=logger,
        )
        cleaner = CleanerUtil(
            data,
            step_details["type"],
            step_details["args"]["columns"],
        )
        data = cleaner.func()
    elif step_details["task"] == "transform":
        step_details["args"]["columns"] = get_columns_names(
            df_columns=data.columns,
            arguments=step_details["args"],
            logger=logger,
        )
        transformer = TransformerUtil(
            data,
            step_details["type"],
        )
        data = transformer.func(**step_details["args"])
    elif step_details["task"] == "math_operation":
        if "result" not in step_details["args"]:
            if len(step_details["args"].keys()) > 1:
                step_details["args"]["result"] = [
                    k for k in step_details["args"].keys() if k != "columns"
                ][0]
            else:
                step_details["args"]["result"] = "result"
        step_details["args"]["columns"] = get_columns_names(
            df_columns=data.columns,
            arguments=step_details["args"],
            logger=logger,
        )
        if len(step_details["args"]["columns"]) < 2:
            logger.info("Math operation requires at least 2 columns. Skipping.")
            return data
        operator = MathOperatorUtil(
            data,
            step_details["type"],
            logger,
            step_details["args"]["columns"],
            step_details["args"]["result"],
        )
        data = operator.func()
    elif step_details["task"] == "analysis":
        analyser = AnalyserUtil(data, step_details["type"], logger)
        data = analyser.func(**step_details["args"])
    return data


def print_df_details(df_dict: dict[pd.DataFrame], df_info_dict: dict[str]) -> str:
    str_output = ""
    for name, df in df_dict.items():
        var_name = name.lower().replace(" ", "_")
        if name in df_info_dict and isinstance(df, pd.DataFrame):
            str_output += f"Dataframe: `{var_name}`\nOutput of `{var_name}.head()`:\n{df.head()}\nOutput of `{var_name}.info()`:\n{df_info_dict[name]}\n"
        else:
            str_output += f"{name}: {df}\n"
    return str_output


class MLAnalysisFactory:
    def __init__(
        self,
        model: LiteLLM,
        data_dict: list[pd.DataFrame],
        data_info_dict: list[str],
        logger: logging.Logger,
        context: str,
    ):
        self.model = model
        self.model.set_model_kwargs(
            model_kwargs={"seed": 123, "temperature": 0.1, "top_p": 0.5}
        )
        self.df_dict = data_dict
        self.df_info_dict = data_info_dict
        self.context = context.strip() + "\n\n" if context != "" else ""
        self.logger = logger

    def _get_analysis_guide(self, user_input: str) -> str:
        output = self.model.run(
            messages=[
                LyzrPromptFactory(
                    name="analysis_guide", prompt_type="system"
                ).get_message(
                    context=self.context,
                ),
                LyzrPromptFactory(
                    name="analysis_guide", prompt_type="user"
                ).get_message(
                    df_details=print_df_details(self.df_dict, self.df_info_dict),
                    question=user_input,
                ),
            ],
            max_tokens=250,
        )
        return output.message.content.strip()

    def _get_analysis_steps_messages_kwargs(self, user_input: str) -> tuple:
        schema = {
            "analysis_df": str,  # Name of the dataframe
            "steps": [
                {
                    "step": 1,
                    "task": "clean_data",
                    "type": "convert_to_datetime",
                    "args": {
                        "columns": ["col1"],
                    },
                },
                {
                    "step": 2,
                    "task": "math_operation",
                    "type": "subtract",
                    "args": {
                        "columns": ["col1", "col2"],
                        "result": "col3",
                    },
                },
                {
                    "step": 3,
                    "task": "analysis",
                    "type": "groupby",
                    "args": {
                        "columns": ["col1"],
                        "agg": "mean",
                        "agg_col": ["col2"],
                    },
                },
            ],
            "output columns": ["col1", "col2", "col3"],
        }
        messages = [
            LyzrPromptFactory(name="analysis_steps", prompt_type="system").get_message(
                schema=schema
            ),
            LyzrPromptFactory(name="analysis_steps", prompt_type="user").get_message(
                df_details=print_df_details(self.df_dict, self.df_info_dict),
                question=user_input,
                context=self.analysis_guide,
            ),
        ]
        return messages, dict(
            response_format={"type": "json_object"},
            max_tokens=2000,
            top_p=1,
        )
        # output = self.model.run(
        #     messages=[
        #         LyzrPromptFactory(
        #             name="analysis_steps", prompt_type="system"
        #         ).get_message(schema=schema),
        #         LyzrPromptFactory(
        #             name="analysis_steps", prompt_type="user"
        #         ).get_message(
        #             df_details=print_df_details(self.df_dict, self.df_info_dict),
        #             question=user_input,
        #             context=self.analysis_guide,
        #         ),
        #     ],
        #     response_format={"type": "json_object"},
        #     max_tokens=2000,
        #     top_p=1,
        # )
        # return output.message.content

    def _get_and_run_analysis(self, user_input) -> tuple:
        messages, kwargs = self._get_analysis_steps_messages_kwargs(user_input)
        for _ in range(5):
            try:
                llm_output = self.model.run(messages=messages, **kwargs).message.content
                messages.append(AssistantMessage(content=llm_output))
                self.logger.info("\nSecond analysis LLM output:\n{}".format(llm_output))
                self.analysis_dict = check_output_format(llm_output, self.logger)
                outputs, data = self.get_analysis_from_steps(self.analysis_dict)
                return outputs, data
            except RecursionError:
                raise RecursionError(
                    "The request could not be completed. Please wait a while and try again."
                )
            except Exception as e:
                if time.time() - self.start_time > 30:
                    raise TimeoutError(
                        "The request could not be completed. Please wait a while and try again."
                    )
                self.logger.info(f"{e.__class__.__name__}: {e}\n")
                self.logger.info("Traceback:\n{}\n".format(traceback.format_exc()))
                messages.append(
                    SystemMessage(
                        content=f"Your response resulted in the following error:\n{e.__class__.__name__}: {e}\n{traceback.format_exc()}\n\nPlease correct your response to prevent this error."
                    )
                )

    def get_analysis_from_steps(self, analysis_dict):
        df_name = analysis_dict["analysis_df"]
        data = self.df_dict[df_name].copy(deep=True)
        outputs = []
        for step_details in analysis_dict["steps"]:
            step_details = validate_output_step_details(step_details, self.logger)
            data = run_analysis_step(data, step_details, self.logger)
            if not isinstance(data, pd.DataFrame):
                outputs.append(data)
                data = self.df_dict[df_name].copy(deep=True)
        return outputs, data

    def _handle_single_output(self, data):
        self.logger.info(f"\nOutput dataframe shape: {data.shape}")
        self.logger.info(f"Output dataframe head:\n{data.head()}\n")

        if not isinstance(data, pd.DataFrame):
            self.output = data
            return

        output_columns = get_columns_names(
            df_columns=data.columns,
            columns=self.analysis_dict["output_columns"],
            logger=self.logger,
        )
        # Return output columns
        if data.size == 1:
            self.output = data[output_columns].values[0]
        else:
            self.output = data[output_columns]

    def _handle_list_output(self, outputs):
        self.logger.info(f"Output list length: {len(outputs)}\n")
        # self.logger.info(f"Output dataframe head: {outputs[0].head()}")

        self.output = outputs

    def run_complete_analysis(self, user_input: str) -> Union[list, pd.DataFrame]:
        # Get analysis steps from LLM using user_input
        self.analysis_guide = self._get_analysis_guide(user_input)
        if (
            self.analysis_guide
            == "The provided dataset is not sufficient to answer the question."
        ):
            return self.analysis_guide
        self.logger.info("First analysis LLM output:\n{}\n".format(self.analysis_guide))

        # Perform analysis using classes from lyzr/data_analyzr/run_analysis.py
        self.start_time = time.time()
        outputs, data = self._get_and_run_analysis(user_input)
        if len(outputs) > 0:
            outputs.append(data)
            self._handle_list_output(outputs)
        else:
            self._handle_single_output(data)

        return self.output


class CleanerUtil:

    def __init__(
        self,
        df: pd.DataFrame,
        cleaning_type: Literal[
            "remove_nulls",
            "convert_to_datetime",
            "convert_to_numeric",
            "convert_to_categorical",
        ],
        columns: list,
    ):
        self.df = df
        self.columns = columns
        self.func = getattr(self, cleaning_type.lower())

    def remove_nulls(self) -> pd.DataFrame:
        self.df.dropna(subset=self.columns, inplace=True)
        return self.df

    def convert_to_datetime(self) -> pd.DataFrame:
        self.df = self.remove_nulls()
        self.df.loc[:, self.columns] = self.df.loc[:, self.columns].apply(
            pd.to_datetime, errors="coerce"
        )
        return self.df.infer_objects()

    def convert_to_numeric(self) -> pd.DataFrame:
        self.df = self.remove_nulls()
        self.df = self._remove_punctuation()
        self.df.loc[:, self.columns] = self.df.loc[:, self.columns].apply(pd.to_numeric)
        return self.df.infer_objects()

    def convert_to_categorical(self) -> pd.DataFrame:
        self.df = self.remove_nulls()
        self.df.loc[:, self.columns] = self.df.loc[:, self.columns].astype("category")
        return self.df.infer_objects()

    def _remove_punctuation(self) -> pd.DataFrame:
        if isinstance(self.df, pd.Series):
            self.df = self.df.to_frame()
        for col in self.columns:
            self.df.loc[:, col] = self.df.loc[:, col].apply(
                CleanerUtil._remove_punctuation_from_string
            )
        return self.df

    @staticmethod
    def _remove_punctuation_from_string(value) -> str:
        if not isinstance(value, str):
            return value
        value = value.strip()
        cleaned = re.sub(r"[^\d.]", "", str(value))
        if cleaned.replace(".", "").isdigit():
            return "-" + cleaned if value[0] == "-" else cleaned
        else:
            return value


class TransformerUtil:
    def __init__(
        self,
        df: pd.DataFrame,
        transform_type: Literal[
            "one_hot_encode",
            "ordinal_encode",
            "scale",
            "extract_time_period",
            "select_indices",
        ],
    ):
        self.df = df
        self.func = getattr(
            self, transform_type.replace("-", "").replace(" ", "_").lower()
        )

    def one_hot_encode(self, columns) -> pd.DataFrame:
        encoded_df = pd.get_dummies(self.df, columns=columns, dtype=float)
        categories = {}
        for col in columns:
            categories[col] = self.df[col].astype("category").cat.categories.to_list()
        return encoded_df

    def ordinal_encode(self, columns) -> pd.DataFrame:
        try:
            from sklearn.preprocessing import OrdinalEncoder
        except ImportError:
            raise DependencyError({"scikit-learn": "scikit-learn==1.4.0"})

        encoder = OrdinalEncoder()
        encoded_df = self.df
        encoded_df.loc[:, columns] = encoder.fit_transform(
            self.df.loc[:, columns].values
        )
        categories = {}
        for col in columns:
            categories[col] = self.df[col].astype("category").cat.categories.to_list()
        return encoded_df

    def scale(self, columns) -> pd.DataFrame:
        try:
            from sklearn.preprocessing import StandardScaler
        except ImportError:
            raise DependencyError({"scikit-learn": "scikit-learn==1.4.0"})

        scaler = StandardScaler()
        scaled_df = self.df
        scaled_df.loc[:, columns] = scaler.fit_transform(self.df.loc[:, columns].values)
        return scaled_df

    def extract_time_period(
        self,
        columns,
        period_to_extract: Literal[
            "week", "month", "year", "day", "hour", "minute", "second", "weekday"
        ],
    ) -> pd.DataFrame:
        columns = columns[0]
        self.df.loc[:, columns] = pd.to_datetime(self.df.loc[:, columns])
        if period_to_extract == "week":
            self.df.loc[:, ["week"]] = (
                pd.to_datetime(self.df.loc[:, columns], errors="coerce")
                .dt.isocalendar()
                .week
            )
        elif period_to_extract == "month":
            self.df.loc[:, ["month"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.month
        elif period_to_extract == "year":
            self.df.loc[:, ["year"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.year
        elif period_to_extract == "day":
            self.df.loc[:, ["day"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.day
        elif period_to_extract == "hour":
            self.df.loc[:, ["hour"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.hour
        elif period_to_extract == "minute":
            self.df.loc[:, ["minute"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.minute
        elif period_to_extract == "second":
            self.df.loc[:, ["second"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.second
        elif period_to_extract == "weekday":
            self.df.loc[:, ["weekday"]] = pd.to_datetime(
                self.df.loc[:, columns], errors="coerce"
            ).dt.weekday
        return self.df

    def select_indices(self, columns, indices) -> pd.DataFrame:
        return self.df.loc[indices, :]


class MathOperatorUtil:
    def __init__(
        self,
        df: pd.DataFrame,
        operator_type: Literal["add", "subtract", "multiply", "divide"],
        logger: logging.Logger,
        columns: list,
        result: str,
    ):
        self.df = df
        self.logger = logger
        self.columns = columns
        self.result = result
        self.df = self.df.dropna(subset=self.columns)
        self.func = getattr(self, operator_type.lower())

    def add(self) -> pd.DataFrame:
        self.df.loc[:, self.result] = self.df.loc[:, self.columns].sum(axis=1)
        return self.df

    def subtract(self) -> pd.DataFrame:
        self.df.loc[:, self.result] = (
            self.df[self.columns[0]] - self.df[self.columns[1]]
        )
        return self.df

    def multiply(self) -> pd.DataFrame:
        self.df.loc[:, self.result] = self.df.loc[:, self.columns].prod(axis=1)
        return self.df

    def divide(self) -> pd.DataFrame:
        self.df = self.df[self.df[self.columns[1]] != 0]
        self.df.loc[:, self.result] = (
            self.df[self.columns[0]] / self.df[self.columns[1]]
        )
        return self.df


class AnalyserUtil:
    def __init__(
        self,
        df: pd.DataFrame,
        analysis_type: Literal[
            "sortvalues",
            "filter",
            "mean",
            "sum",
            "cumsum",
            "groupby",
            "correlation",
            "regression",
        ],
        logger: logging.Logger,
    ):
        self.df = df
        self.logger = logger
        self.func = getattr(self, analysis_type.lower())

    def sortvalues(
        self, columns: list, ascending: Optional[Any] = True
    ) -> pd.DataFrame:
        if isinstance(ascending, list) and len(ascending) == len(columns):
            for col, asc in zip(columns, ascending):
                self.df = self._sorter([col], asc)
        elif isinstance(ascending, str) or isinstance(ascending, bool):
            self.df = self._sorter(columns, ascending)
        else:
            self.logger.warning(
                "Invalid value provided for ascending. Defaulting to True."
            )
            self.df = self._sorter(columns, True)
        return self.df

    def _sorter(self, columns: list, ascending: Any) -> pd.DataFrame:
        if isinstance(ascending, str):
            if ascending.lower() == "true":
                ascending = True
            elif ascending.lower() == "false":
                ascending = False
            else:
                self.logger.warning(
                    f"Invalid value provided for ascending: {ascending}. Defaulting to True."
                )
                ascending = True
        if not isinstance(ascending, bool):
            self.logger.warning(
                f"Invalid value type provided for ascending: {type(ascending)}. Defaulting to True."
            )
            ascending = True
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        return self.df.sort_values(columns, ascending=ascending)

    def filter(
        self,
        columns: list[str],
        values: list[Any],
        relations: list[
            Literal[
                "lessthan",
                "greaterthan",
                "lessthanorequalto",
                "greaterthanorequalto",
                "equalto",
                "notequalto",
                "startswith",
                "endswith",
                "contains",
            ]
        ],
    ) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )

        if len(columns) != len(values) and len(values) == 1:
            values = values * len(columns)
        if len(columns) != len(relations) and len(relations) == 1:
            relations = relations * len(columns)
        if len(columns) != len(values) or len(columns) != len(relations):
            self.logger.warning(
                "Invalid number of columns, values or relations provided. Returning original dataframe."
            )
            return self.df

        for i in range(len(columns)):
            col, val, rel = columns[i], values[i], relations[i]
            if rel == "startswith":
                self.df = self.df[self.df[col].str.startswith(str(val))]
            elif rel == "endswith":
                self.df = self.df[self.df[col].str.endswith(str(val))]
            elif rel == "contains":
                self.df = self.df[self.df[col].str.contains(str(val))]
            elif rel == "lessthan":
                self.df = self.df[self.df[col] < val]
            elif rel == "greaterthan":
                self.df = self.df[self.df[col] > val]
            elif rel == "lessthanorequalto":
                self.df = self.df[self.df[col] <= val]
            elif rel == "greaterthanorequalto":
                self.df = self.df[self.df[col] >= val]
            elif rel == "equalto":
                self.df = self.df[self.df[col] == val]
            elif rel == "notequalto":
                self.df = self.df[self.df[col] != val]
            else:
                self.logger.warning(f"Invalid relation provided: {rel}. Skipping.")
        return self.df

    def mean(self, columns: list, result: Optional[str] = None) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        if result is not None:
            self.df.loc[:, result] = self.df.loc[:, columns].mean()
            return self.df
        return self.df.loc[:, columns].mean()

    def sum(self, columns: list, result: Optional[str] = None) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        if result is not None:
            self.df.loc[:, result] = self.df.loc[:, columns].sum(axis=1)
            return self.df
        return self.df.loc[:, columns].sum()

    def cumsum(self, columns: list, result: Optional[str] = None) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        if result is not None:
            self.df.loc[:, result] = self.df.loc[:, columns].cumsum()
            return self.df
        return self.df.loc[:, columns].cumsum()

    def groupby(
        self,
        columns: list,
        agg: Union[str, list],
        result: Optional[str] = None,
        agg_col: Optional[list] = None,
    ) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        if len(columns) == 0:
            self.logger.warning(
                "No valid columns provided. Returning original dataframe."
            )
            return self.df
        if agg_col is None:
            agg_col = self.df.columns.to_list()
        agg_col = get_columns_names(
            df_columns=self.df.columns, columns=agg_col, logger=self.logger
        )
        if result is not None:
            self.df.loc[:, result] = self.df.groupby(columns)[agg_col].agg(agg)
            return self.df
        return self.df.groupby(columns)[agg_col].agg(agg).reset_index()

    def correlation(
        self,
        columns: list,
        method: Optional[str] = "pearson",
        result: Optional[str] = None,
    ) -> pd.DataFrame:
        columns = get_columns_names(
            df_columns=self.df.columns, columns=columns, logger=self.logger
        )
        if result is not None:
            self.df.loc[:, result] = self.df.loc[:, columns].corr(method=method)
            return self.df
        return self.df.loc[:, columns].corr(method=method)

    def regression(self, x: list, y: list) -> pd.DataFrame:
        try:
            from sklearn.linear_model import LinearRegression
        except ImportError:
            raise DependencyError({"scikit-learn": "scikit-learn==1.4.0"})

        x = get_columns_names(df_columns=self.df.columns, columns=x, logger=self.logger)
        y = get_columns_names(df_columns=self.df.columns, columns=y, logger=self.logger)
        model = LinearRegression()
        model.fit(self.df.loc[:, x], self.df.loc[:, y])
        return model

    def forecast(
        self,
        time_column: str,
        y_column: str,
        end: Optional[str] = None,
        steps: Optional[int] = None,
    ):
        try:
            import pmdarima as pm
        except ImportError:
            raise DependencyError({"pmdarima": "pmdarima==2.0.4"})

        time_column = get_columns_names(
            df_columns=self.df.columns, columns=[time_column], logger=self.logger
        )[0]
        y_column = get_columns_names(
            df_columns=self.df.columns, columns=[y_column], logger=self.logger
        )[0]
        data = self.df.loc[
            self.df[time_column].drop_duplicates().sort_values().index,
            [time_column, y_column],
        ]
        data = data.set_index([time_column])
        data.index = pd.DatetimeIndex(pd.to_datetime(data.index))
        model = pm.auto_arima(data[y_column])
        if end is not None:
            steps = len(
                pd.date_range(
                    start=data.index[-1],
                    end=pd.to_datetime(end),
                    freq=data.index.inferred_freq,
                )
            )
        y_pred = model.predict(n_periods=steps)
        for datetime in y_pred.index:
            data.loc[datetime, y_column] = y_pred[datetime]
        return data



# standard library imports
import os
import time
import logging
import traceback
from typing import Union

# third-party imports
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# local imports
from lyzr.base.base import SystemMessage, AssistantMessage
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.llm import LiteLLM
from lyzr.data_analyzr.data_connector import DataConnector
from lyzr.data_analyzr.txt_to_sql_utils import TxttoSQLFactory
from lyzr.data_analyzr.output_handler import check_output_format
from lyzr.data_analyzr.ml_analysis_utils import MLAnalysisFactory
from lyzr.data_analyzr.utils import (
    convert_to_numeric,
    flatten_list,
    get_columns_names,
    format_df_with_describe,
)


class PlotFactory:

    def __init__(
        self,
        model: LiteLLM,
        logger: logging.Logger,
        plot_context: str,
        plot_path: str,
        df_dict: dict,
        database_connector: DataConnector,
        analysis_type: str,
        analyzer: Union[MLAnalysisFactory, TxttoSQLFactory],
        analysis_output: Union[pd.DataFrame, list],
    ):
        self.model = model
        self.model.set_model_kwargs(
            model_kwargs=dict(seed=123, temperature=0.1, top_p=0.5)
        )
        if not isinstance(analysis_output, pd.DataFrame) or analysis_output.size < 2:
            self.use_analysis = True
            self.analysis_type = analysis_type
            self.analyzer = analyzer
            self.df_dict = df_dict
            self.database_connector = database_connector
            self.analysis_output = analysis_output
        else:
            self.use_analysis = False
            self.analysis_output = analysis_output
            self.df_dict = None
            self.database_connector = None
            self.analyzer = None
            self.analysis_type = None
        self.logger = logger
        self.context = (
            plot_context.strip() + "\n\n" if plot_context.strip() != "" else ""
        )
        self.plotting_library = "matplotlib"
        self.output_format = "png"
        self.plot_path = self._handle_plotpath(plot_path)

    def _handle_plotpath(self, plot_path) -> str:
        plot_path = PlotFactory._fix_plotpath(plot_path)
        try:
            open(plot_path, "w").close()
            return plot_path
        except Exception:
            self.logger.warning(
                f'Incorrect path for plot image provided: {self.plot_path}. Defaulting to "generated_plots/plot.png".'
            )
            return self._handle_plotpath("generated_plots/plot.png")

    @staticmethod
    def _fix_plotpath(plot_path: str) -> str:
        if os.path.isdir(plot_path):
            plot_path = os.path.join(plot_path, "plot.png")
        if os.path.splitext(plot_path)[1] != ".png":
            plot_path = os.path.splitext(plot_path)[0] + ".png"
        dir_path = os.path.dirname(plot_path)
        if dir_path.strip() != "":
            os.makedirs(dir_path, exist_ok=True)
        return plot_path

    def _get_plotting_guide(self, user_input: str) -> str:
        plotting_guide_sections = ["context", "external_context"]
        if self.use_analysis and (self.analysis_type == "ml"):
            plotting_guide_sections.append("task_with_analysis")
            df_details = format_df_with_describe(self.df_dict)
        else:
            plotting_guide_sections.append("task_no_analysis")
            df_details = format_df_with_describe(self.analysis_output)
        output = self.model.run(
            messages=[
                LyzrPromptFactory(
                    name="plotting_guide", prompt_type="system"
                ).get_message(
                    use_sections=plotting_guide_sections,
                    context=self.context,
                    plotting_lib=self.plotting_library,
                ),
                LyzrPromptFactory(
                    name="plotting_guide", prompt_type="user"
                ).get_message(
                    df_details=df_details,
                    question=user_input,
                ),
            ],
            max_tokens=500,
        )
        return output.message.content.strip()

    def _get_plotting_steps_no_analysis_messages(self, user_input: str) -> list:
        schema = {
            "figsize": (int, int),
            "subplots": (int, int),
            "title": str,
            "plots": [
                {
                    "subplot": (int, int),
                    "plot_type": "line",  # line, bar, barh, scatter, hist
                    "x": str,
                    "y": str,
                    "args": {
                        "xlabel": str,
                        "ylabel": str,
                        "color": str,
                        "linestyle": str,  # for line plots
                    },
                },
                {
                    "subplot": (int, int),
                    "plot_type": "bar",  # line, bar, barh, scatter, hist
                    "x": str,
                    "y": str,
                    "args": {
                        "xlabel": str,
                        "ylabel": str,
                        "color": str,
                        "stacked": bool,  # for bar plots
                    },
                },
            ],
        }
        messages = [
            LyzrPromptFactory(name="plotting_steps", prompt_type="system").get_message(
                use_sections=["context", "task_no_analysis", "closing"],
                plotting_lib=self.plotting_library,
                schema=schema,
            ),
            LyzrPromptFactory(name="plotting_steps", prompt_type="user").get_message(
                question=user_input,
                df_details=format_df_with_describe(self.analysis_output),
                guide=self.plotting_guide,
            ),
        ]
        return messages

    def _get_plotting_steps_with_analysis_messages(self, user_input: str) -> list:
        schema = {
            "preprocess": {
                "analysis_df": str,
                "steps": [
                    {
                        "step": 1,
                        "task": "clean_data",
                        "type": "convert_to_datetime",
                        "args": {
                            "columns": ["col1"],
                        },
                    },
                    {
                        "step": 2,
                        "task": "math_operation",
                        "type": "subtract",
                        "args": {
                            "columns": ["col1", "col2"],
                            "result": "col3",
                        },
                    },
                    {
                        "step": 3,
                        "task": "analysis",
                        "type": "groupby",
                        "args": {
                            "columns": ["col1"],
                            "agg": "mean",
                            "agg_col": ["col2"],
                        },
                    },
                ],
            },
            "plot": {
                "figsize": (int, int),
                "subplots": (int, int),
                "title": str,
                "plots": [
                    {
                        "subplot": (int, int),
                        "plot_type": "line",  # line, bar, barh, scatter, hist
                        "x": str,
                        "y": str,
                        "args": {
                            "xlabel": str,
                            "ylabel": str,
                            "color": str,
                            "linestyle": str,  # for line plots
                        },
                    },
                    {
                        "subplot": (int, int),
                        "plot_type": "bar",  # line, bar, barh, scatter, hist
                        "x": str,
                        "y": str,
                        "args": {
                            "xlabel": str,
                            "ylabel": str,
                            "color": str,
                            "stacked": bool,  # for bar plots
                        },
                    },
                ],
            },
        }
        messages = [
            LyzrPromptFactory(name="plotting_steps", prompt_type="system").get_message(
                use_sections=["context", "task_with_analysis", "closing"],
                plotting_lib=self.plotting_library,
                schema=schema,
            ),
            LyzrPromptFactory(name="plotting_steps", prompt_type="user").get_message(
                question=user_input,
                df_details=format_df_with_describe(self.analysis_output),
                guide=self.plotting_guide,
            ),
        ]
        return messages

    def retry_plotting_steps(
        self,
        messages: list,
    ) -> bool:
        for _ in range(5):
            try:
                llm_output = self.model.run(
                    messages=messages,
                    response_format={"type": "json_object"},
                    max_tokens=2000,
                    top_p=1,
                ).message.content
                messages.append(AssistantMessage(content=llm_output))
                self.all_steps = check_output_format(llm_output, self.logger, "plot")
                self.logger.info(f"\nPlotting steps recieved:\n{self.all_steps}")
                self.preprocess_steps = self.all_steps.get("preprocess", None)
                if self.preprocess_steps is not None and len(self.preprocess_steps) > 0:
                    _, self.analysis_output = self.analyzer.get_analysis_from_steps(
                        self.preprocess_steps
                    )
                    self.plotting_steps = self.all_steps["plot"]
                else:
                    self.plotting_steps = self.all_steps
                self.fig = self._create_plot(self.analysis_output)
                return True
            except RecursionError:
                plt.close("all")
                raise RecursionError(
                    "The request could not be completed. Please wait a while and try again."
                )
            except Exception as e:
                plt.close("all")
                if time.time() - self.start_time > 30:
                    raise TimeoutError(
                        "The request could not be completed. Please wait a while and try again."
                    )
                self.logger.info(f"{e.__class__.__name__}: {e}\n")
                self.logger.info("Traceback:\n{}\n".format(traceback.format_exc()))
                messages.append(
                    SystemMessage(
                        content=f"Your response resulted in the following error:\n{e.__class__.__name__}: {e}\n{traceback.format_exc()}\n\nPlease correct your response to prevent this error."
                    )
                )
        return False

    def _set_args_stacked(self, args: dict) -> bool:
        stacked = args.get("stacked", False)
        if isinstance(stacked, str):
            if stacked.lower() == "true":
                stacked = True
            elif stacked.lower() == "false":
                stacked = False
            else:
                stacked = False
        if not isinstance(stacked, bool):
            self.logger.warning(
                f"Invalid value type provided for stacked: {type(args['stacked'])}. Defaulting to False."
            )
            stacked = False
        return stacked

    def _get_bar_df(self, df: pd.DataFrame, columns: list) -> pd.DataFrame:
        n_bars = 25
        if df[columns].shape[0] > n_bars:
            self.logger.warning(
                f"\nToo many bars given. Plotting only the top {n_bars} bars."
            )
            num_cols = df[columns].select_dtypes(include=np.number).columns.tolist()
            df_bar = df[columns].sort_values(by=num_cols, ascending=False).head(n_bars)
        else:
            df_bar = df[columns]
        return df_bar

    def _plot_subplot(
        self, plot_type: str, axes: np.ndarray, args: dict, df: pd.DataFrame, plot: dict
    ) -> None:
        if "x" in plot and isinstance(plot["x"], list):
            plot["x"] = plot["x"][0]
        if "y" in plot and isinstance(plot["y"], list):
            plot["y"] = plot["y"][0]
        columns = get_columns_names(
            df_columns=df.columns,
            columns=list(
                flatten_list([plot.get("x", []), plot.get("y", []), plot.get("by", [])])
            ),
            logger=self.logger,
        )
        df = convert_to_numeric(df, columns=columns).infer_objects()

        if plot_type == "line":
            self.logger.info(f"\nDF to be plot:\n{df.head()}\n")
            df.plot.line(
                x=plot.get("x"),
                y=plot.get("y"),
                ax=axes,
                **args,
            )
        elif plot_type == "bar":
            args["stacked"] = self._set_args_stacked(args)
            df_bar = self._get_bar_df(df, columns)
            self.logger.info(f"\nDF to be plot:\n{df_bar.head()}\n")
            df_bar.plot.bar(
                x=plot.get("x"),
                y=plot.get("y"),
                ax=axes,
                **args,
            )
        elif plot_type == "barh":
            args["stacked"] = self._set_args_stacked(args)
            df_bar = self._get_bar_df(df, columns)
            self.logger.info(f"\nDF to be plot:\n{df_bar.head()}\n")
            df_bar.plot.barh(
                x=plot.get("x"),
                y=plot.get("y"),
                ax=axes,
                **args,
            )
        elif plot_type == "scatter":
            self.logger.info(f"\nDF to be plot:\n{df.head()}\n")
            df.plot.scatter(
                x=plot.get("x"),
                y=plot.get("y"),
                ax=axes,
                **args,
            )
        elif plot_type == "hist":
            df_hist = df[plot.get("by")]
            self.logger.info(f"\nDF to be plot:\n{df_hist.head()}\n")
            df_hist.plot.hist(
                ax=axes,
                **args,
            )
        else:
            raise ValueError(f"Invalid plot_type received: {plot_type}.")
        xlabel = plot.get("args", {}).get("xlabel")
        ylabel = plot.get("args", {}).get("ylabel")
        if xlabel:
            axes.set_xlabel(xlabel)
        if ylabel:
            axes.set_ylabel(ylabel)

    def _create_plot(self, df: pd.DataFrame) -> plt.Figure:
        nrows, ncols = self.plotting_steps.get("subplots", (1, 1))
        fig, ax = plt.subplots(
            figsize=self.plotting_steps.get("figsize", (10, 10)),
            nrows=nrows,
            ncols=ncols,
        )
        if not isinstance(ax, np.ndarray):
            ax = np.array([ax])
        ax = ax.reshape((nrows, ncols))

        fig.suptitle(self.plotting_steps.get("title", "Plot"))
        for plot in self.plotting_steps.get("plots", []):
            self.logger.info(f"\nPlotting: {plot}")
            plot_type = plot.get("plot_type", "line")
            axes = ax[
                plot.get("subplot", (1, 1))[0] - 1, plot.get("subplot", (1, 1))[1] - 1
            ]
            args = plot.get("args", {})
            self._plot_subplot(plot_type, axes, args, df, plot)
        return fig

    def get_visualisation_image(self) -> str:
        plt.tight_layout()
        if not PlotFactory._savefig(self.fig, self.plot_path, self.logger):
            self.logger.error(
                f"Error saving plot at: {self.plot_path}. Plot not saved. Displaying plot instead. Access the plot using `.fig` attribute."
            )
            plt.show()
        else:
            self.logger.info(f"\nPlot saved at: {self.plot_path}\n")
            plt.close(self.fig)
        return self.plot_path

    @staticmethod
    def _savefig(fig, path, logger):
        try:
            dir_path = os.path.dirname(path)
            if dir_path.strip() != "":
                os.makedirs(dir_path, exist_ok=True)
            fig.savefig(path)
            return True
        except Exception:
            logger.error(
                f"Error saving plot at: {path}. Trying to save at default location: 'generated_plots/plot.png'."
            )
            PlotFactory._savefig(fig, "generated_plots/plot.png", logger)
        return False

    def get_visualisation(self, user_input: str) -> str:
        self.start_time = time.time()
        self.plotting_guide = self._get_plotting_guide(user_input)
        self.logger.info(f"\nPlotting guide recieved:\n{self.plotting_guide}")
        if self.use_analysis:
            if self.analysis_type == "ml":
                messages = self._get_plotting_steps_with_analysis_messages(user_input)
            elif self.analysis_type == "sql":
                self.analysis_output = self.analyzer.run_analysis_for_plotting(
                    user_input=user_input
                )
                messages = self._get_plotting_steps_no_analysis_messages(user_input)
        else:
            messages = self._get_plotting_steps_no_analysis_messages(user_input)
        if self.retry_plotting_steps(messages):
            self.get_visualisation_image()
            return self.plot_path



# standard library imports
import time
import logging
import traceback

# local imports
from lyzr.base.prompt import LyzrPromptFactory
from lyzr.base.errors import MissingValueError
from lyzr.base.llm import LiteLLM
from lyzr.base.base import ChatMessage, UserMessage, SystemMessage
from lyzr.data_analyzr.output_handler import extract_sql
from lyzr.data_analyzr.db_connector import DatabaseConnector
from lyzr.data_analyzr.vector_store_utils import ChromaDBVectorStore
from lyzr.data_analyzr.utils import run_n_times


class TxttoSQLFactory:

    def __init__(
        self,
        model: LiteLLM,
        db_connector: DatabaseConnector,
        logger: logging.Logger,
        context: str,
        vector_store: ChromaDBVectorStore,
    ):
        self.model = model
        self.model.set_model_kwargs(
            model_kwargs=dict(seed=123, temperature=0.1, top_p=0.5)
        )
        self.context = context.strip() + "\n\n" if context.strip() != "" else ""
        self.logger = logger
        self.connector = db_connector
        self.vector_store = vector_store
        if self.vector_store is None:
            raise MissingValueError("vector_store")
        self.analysis_output, self.analysis_guide = None, None

    def run_complete_analysis(
        self,
        user_input: str,
        auto_train: bool = True,
        **kwargs,
    ):
        start_time = time.time()
        question_sql_list = self.vector_store.get_similar_question_sql(
            user_input, **kwargs
        )
        ddl_list = self.vector_store.get_related_ddl(user_input, **kwargs)
        doc_list = self.vector_store.get_related_documentation(user_input, **kwargs)
        messages = self._get_sql_prompt(
            user_input=user_input,
            question_sql_list=question_sql_list,
            ddl_list=ddl_list,
            doc_list=doc_list,
        )
        # print(messages)
        for _ in range(3):
            try:
                # SQL Generation
                self.analysis_guide = extract_sql(
                    self.model.run(messages=messages, **kwargs).message.content,
                    self.logger,
                )
                self.logger.info(f"Analysis Guide:\n{self.analysis_guide}\n")
                # SQL Execution
                self.analysis_output = self.connector.run_sql(self.analysis_guide)
                self.logger.info(f"Analysis Output:\n{self.analysis_output}\n")
                break
            except RecursionError:
                raise RecursionError(
                    "The request could not be completed. Please wait a while and try again."
                )
            except Exception as e:
                if time.time() - start_time > 30:
                    raise TimeoutError(
                        "The request could not be completed. Please wait a while and try again."
                    )
                self.logger.info(f"{e.__class__.__name__}: {e}\n")
                self.logger.info("Traceback:\n{}\n".format(traceback.format_exc()))
                messages.append(
                    SystemMessage(
                        content=f"Your response resulted in the following error:\n{e.__class__.__name__}: {e}\n{traceback.format_exc()}\n\nPlease correct your response to prevent this error."
                    )
                )
        # Auto-training
        if user_input is None or user_input.strip() == "":
            user_input = self._generate_question(self.analysis_guide)
        if (
            auto_train
            and self.analysis_output is not None
            and len(self.analysis_output) > 0
            and self.analysis_guide is not None
            and self.analysis_guide.strip() != ""
        ):
            self.logger.info("Saving data for next training round\n")
            self.vector_store.add_training_plan(
                question=user_input, sql=self.analysis_guide
            )
        return self.analysis_output

    def run_analysis_for_plotting(self, user_input: str, **kwargs):
        start_time = time.time()
        question_sql_list = self.vector_store.get_similar_question_sql(
            user_input, **kwargs
        )
        ddl_list = self.vector_store.get_related_ddl(user_input, **kwargs)
        doc_list = self.vector_store.get_related_documentation(user_input, **kwargs)
        messages = self._get_sql_prompt(
            user_input=user_input,
            question_sql_list=question_sql_list,
            ddl_list=ddl_list,
            doc_list=doc_list,
            for_plotting=True,
        )
        for _ in range(3):
            try:
                sql_query = extract_sql(
                    self.model.run(messages=messages, **kwargs).message.content,
                    self.logger,
                )
                plot_df = self.connector.run_sql(sql_query)
                return plot_df
            except RecursionError:
                raise RecursionError(
                    "The request could not be completed. Please wait a while and try again."
                )
            except Exception as e:
                if time.time() - start_time > 30:
                    raise TimeoutError(
                        "The request could not be completed. Please wait a while and try again."
                    )
                self.logger.info(f"{e.__class__.__name__}: {e}\n")
                self.logger.info("Traceback:\n{}\n".format(traceback.format_exc()))
                messages.append(
                    SystemMessage(
                        content=f"Your response resulted in the following error:\n{e.__class__.__name__}: {e}\n{traceback.format_exc()}\n\nPlease correct your response to prevent this error."
                    )
                )

    def _generate_sql(
        self, user_input: str, for_plotting: bool = False, **kwargs
    ) -> str:
        question_sql_list = self.vector_store.get_similar_question_sql(
            user_input, **kwargs
        )
        ddl_list = self.vector_store.get_related_ddl(user_input, **kwargs)
        doc_list = self.vector_store.get_related_documentation(user_input, **kwargs)
        # get response from LLM in llm_response
        llm_response = self.model.run(
            messages=self._get_sql_prompt(
                user_input=user_input,
                question_sql_list=question_sql_list,
                ddl_list=ddl_list,
                doc_list=doc_list,
                for_plotting=for_plotting,
            ),
            **kwargs,
        ).message.content
        return llm_response

    def _generate_question(self, sql: str, **kwargs) -> str:
        kwargs["max_tokens"] = 500
        output = self.model.run(
            messages=[
                LyzrPromptFactory(
                    name="sql_question_gen", prompt_type="system"
                ).get_message(),
                UserMessage(content=sql),
            ],
            **kwargs,
        )
        return output.message.content

    def _get_sql_prompt(
        self,
        user_input: str,
        question_sql_list: list,
        ddl_list: list,
        doc_list: list,
        for_plotting: bool = False,
    ) -> list[ChatMessage]:
        system_message_sections = ["context", "external_context"]
        system_message_sections.append("task" if not for_plotting else "plotting_task")
        system_message_dict = {"context": self.context}

        if len(ddl_list) > 0:
            system_message_sections.append("ddl_addition_text")
            system_message_dict["ddl"] = ""
            for ddl_item in ddl_list:
                system_message_dict["ddl"] += f"{ddl_item}\n"

        if len(doc_list) > 0:
            system_message_sections.append("doc_addition_text")
            system_message_dict["doc"] = ""
            for doc_item in doc_list:
                system_message_dict["doc"] += f"{doc_item}\n"

        system_message_sections.append("closing")
        messages = [
            LyzrPromptFactory(
                name="txt_to_sql",
                prompt_type="system",
                use_sections=system_message_sections,
            ).get_message(**system_message_dict)
        ]

        for example in question_sql_list:
            if example is not None:
                if "question" in example and "sql" in example:
                    messages.append(UserMessage(content=example["question"]))
                    messages.append(SystemMessage(content=example["sql"]))

        messages.append(UserMessage(content=user_input))
        return messages



from setuptools import setup, find_packages

setup(
    name="lyzr",
    version="0.1.32",
    author="lyzr",
    description="",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="",
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8.1, <3.12",
    install_requires=[
        "asyncio",
        "nest_asyncio",
        "openai==1.3.4",
        "litellm==1.2.0",
        "llama-index==0.9.4",
        "langchain==0.0.339",
        "python-dotenv>=1.0.0",
        "beautifulsoup4==4.12.2",
        "pandas==2.0.2",
        "weaviate-client==3.25.3",
        "llmsherpa",
        "matplotlib==3.8.2",
    ],
    extras_require={
        "data-analyzr": [
            "scikit-learn==1.4.0",
            "statsmodels==0.14.1",
            "chromadb==0.4.22",
            "tabulate==0.9.0",
            "pmdarima==2.0.4",
            "openpyxl==3.1.2",
            "redshift_connector==2.0.918",
            "mysql-connector-python==8.2.0",
            "psycopg2-binary==2.9.9",
            "snowflake-connector-python==3.6.0",
        ],
    },
)
